import numpy as np
from statistics import mode
from .supervised import DecisionTree,RegresiLinear
from ..datamanipulasi.fold import train_val_split

class Bagging:
    """
    Kelas Bagging digunakan untuk melakukan ensemble learning dengan metode Bagging.
    Bagging adalah teknik ensemble learning yang menggabungkan beberapa model yang sama
    jenisnya untuk meningkatkan performa prediksi.

    Parameters:
    - model: objek model yang akan digunakan untuk melakukan prediksi.
    - n_model: jumlah model yang akan digunakan dalam ensemble (default: 5).
    - random_state: seed untuk mengontrol randomness dalam pemilihan data (default: None).

    Methods:
    - fit(x, y): Melakukan training pada model dengan data x dan label y.
    - fit_predict(x, y): Melakukan training pada model dengan data x dan label y, dan mengembalikan hasil prediksi.
    - score_accuracy(y_pred, y_true): Menghitung akurasi prediksi berdasarkan hasil prediksi (y_pred) dan label sebenarnya (y_true).
    """

    def __init__(self, model, n_model: int = 5, random_state: int = None):
        self.__model = model
        self.__label: np.ndarray = None
        self.__random_state = random_state
        self.__n_model = n_model
    
    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Melakukan training pada model dengan data x dan label y.

        Parameters:
        - x: array numpy yang berisi data training.
        - y: array numpy yang berisi label data training.
        """
        if self.__random_state:
            np.random.seed(self.__random_state)
        record_x = [i for i in range(len(x))]
        predict_model = []

        for i in range(self.__n_model):
            data_x = []
            data_y = []
            for _ in range(len(x)):
                rd_c = np.random.choice(record_x, 1, replace=True)
                data_x.append(x[rd_c])
                data_y.append(int(y[rd_c]))
            
            data_x = np.array(data_x).reshape(x.shape)
            data_y = np.array(data_y)
            obj = self.__model
            obj.fit(data_x, data_y)
            predict_model.append(obj)
            pred = obj.predict(x)
            print(
                f"model-{i+1} akurasi : {self.score_accuracy(pred, y)}"
            )

        
        predict = []
        for data in x:
            temp_predict = []
            for m in predict_model:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))


        
        self.__label = np.array(predict)
    
    def fit_predict(self, x: np.ndarray, y: np.ndarray):
        """
        Melakukan training pada model dengan data x dan label y, dan mengembalikan hasil prediksi.

        Parameters:
        - x: array numpy yang berisi data training.
        - y: array numpy yang berisi label data training.

        Returns:
        - label hasil prediksi.
        """
        self.fit(x, y)
        return self.__label
    
    def score_accuracy(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Menghitung akurasi prediksi berdasarkan hasil prediksi (y_pred) dan label sebenarnya (y_true).

        Parameters:
        - y_pred: array numpy yang berisi hasil prediksi.
        - y_true: array numpy yang berisi label sebenarnya.

        Returns:
        - akurasi prediksi.
        """
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true / len(y_pred)


class Boosting:
    """
    Kelas Boosting digunakan untuk melakukan ensemble learning dengan metode Boosting.
    """

    def __init__(self,model, n_model:int = 5,random_state:int=None):
        """
        Inisialisasi objek Boosting.

        Parameters:
        - model: Objek model yang akan digunakan dalam ensemble learning.
        - n_model: Jumlah model yang akan digunakan dalam ensemble learning. Default: 5.
        - random_state: Seed untuk mengatur random state. Default: None.
        """
        self.__model = model
        self.__n_model = n_model
        self.__random_state = random_state
        self.__label = None
    
    def fit(self,x:np.ndarray,y:np.ndarray):
        """
        Melakukan fitting model dengan data training.

        Parameters:
        - x: Array numpy berisi data training.
        - y: Array numpy berisi label data training.
        """
        sample_weight = np.ones(len(x))/len(x)

        if self.__random_state :
            np.random.seed(self.__random_state)
        predict_model = []
        weight_model = []
        final_prediksi = None

        for i in range(self.__n_model):
            selected_indices = np.random.choice(len(x), size=len(x), p=sample_weight)
            data_x = [x[i] for i in selected_indices]
            data_y = [y[i] for i in selected_indices]
            data_x = np.array(data_x)
            data_y = np.array(data_y)
            obj = self.__model
            obj.fit(data_x,data_y)
            predict_model.append(obj)
            prediksi = obj.predict(x)
            error = np.sum(sample_weight * (prediksi != y))
            model_weight = 0.5 * np.log((1 - error) / error)
            weight_model.append(model_weight)
            sample_weight *= np.exp(-model_weight * y * prediksi)
            sample_weight /= np.sum(sample_weight)
            if final_prediksi == None:
                final_prediksi = model_weight*prediksi
            else:
                final_prediksi += model_weight*prediksi
            
            print(
                f"model-{i+1} akurasi : {self.score_accuracy(prediksi,y)}"
            )
            
        self.__label = np.argmax(final_prediksi,axis=1)
    
    def fit_predict(self,x:np.ndarray,y:np.ndarray):
        """
        Melakukan fitting model dengan data training dan mengembalikan label hasil prediksi.

        Parameters:
        - x: Array numpy berisi data training.
        - y: Array numpy berisi label data training.

        Returns:
        - Array numpy berisi label hasil prediksi.
        """
        self.fit(x,y)
        return self.__label

    def score_accuracy(self,y_pred:np.ndarray,y_true:np.ndarray):
        """
        Menghitung akurasi prediksi.

        Parameters:
        - y_pred: Array numpy berisi label hasil prediksi.
        - y_true: Array numpy berisi label sebenarnya.

        Returns:
        - Nilai akurasi prediksi.
        """
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true/len(y_pred)

class RandomForest:
    """
    Implementasi kelas RandomForest untuk melakukan pembelajaran dan prediksi menggunakan metode Random Forest.

    Parameters:
    - max_feature (int): Jumlah fitur maksimum yang akan dipertimbangkan saat membangun setiap pohon keputusan dalam ensemble.
    - min_fitur (int, optional): Jumlah fitur minimum yang akan dipertimbangkan saat membangun setiap pohon keputusan dalam ensemble. Default: 0.
    - max_depth (int): Kedalaman maksimum setiap pohon keputusan dalam ensemble.
    - random_state (int, optional): Seed untuk menghasilkan angka acak. Default: None.
    - n_tree (int): Jumlah pohon keputusan dalam ensemble.

    Attributes:
    - label: Label hasil prediksi setelah melakukan pembelajaran.
    - tree: Daftar pohon keputusan dalam ensemble.

    Methods:
    - fit(x, y, nama_fitur=None): Melakukan pembelajaran pada data x dan label y.
    - fit_predict(x, y, nama_fitur=None): Melakukan pembelajaran pada data x dan label y, serta mengembalikan hasil prediksi.
    - predict(x): Melakukan prediksi pada data x.
    - score_accuracy(y_pred, y_true): Menghitung akurasi prediksi berdasarkan hasil prediksi (y_pred) dan label sebenarnya (y_true).
    """

    def __init__(self, max_feature: int, min_fitur: int = 0, max_depth: int = 2, random_state: int = None, n_tree: int = 3):
        self.__max_depth = max_depth
        self.__random_state = random_state
        self.__n_tree = n_tree
        self.__tree = []
        self.__label = None
        self.__max_feature = max_feature
        self.__min_fitur = min_fitur
    
    @property
    def label(self):
        """
        Getter untuk atribut label.

        Returns:
        - label: Label hasil prediksi setelah melakukan pembelajaran.
        """
        pass

    @label.getter
    def label__(self):
        return self.__label

    @property
    def tree(self):
        """
        Getter untuk atribut tree.

        Returns:
        - tree: Daftar pohon keputusan dalam ensemble.
        """
        pass

    @tree.getter
    def tree__(self):
        return self.__tree
    
    def fit(self, x: np.ndarray, y: np.ndarray, nama_fitur: list = None):
        """
        Melakukan pembelajaran pada data x dan label y menggunakan metode Random Forest.

        Parameters:
        - x (np.ndarray): Data fitur untuk pembelajaran.
        - y (np.ndarray): Label yang sesuai dengan data fitur.
        - nama_fitur (list, optional): Nama fitur untuk keperluan informasi tambahan. Default: None.
        """
        if self.__random_state:
            np.random.seed(self.__random_state)
        record_x = [i for i in range(len(x))]

        for i in range(self.__n_tree):
            # boosting data
            data_x = []
            data_y = []
            data_terpilih = []
            for _ in range(len(x)):
                rd_c = np.random.choice(record_x, 1, replace=True)
                
                data_x.append(x[rd_c])
                data_y.append(int(y[rd_c]))
            data_x = np.array(data_x).reshape(x.shape)
            data_y = np.array(data_y)
            obj = DecisionTree(self.__max_depth)
            obj.fit(data_x, data_y, self.__max_feature, self.__min_fitur)
            self.__tree.append(obj)
            data_x_oob = np.array([x[i] for i in range(len(x)) if i not in data_terpilih])
            data_y_oob = np.array([y[i] for i in range(len(x)) if i not in data_terpilih])
            
            pred = obj.predict(data_x_oob)
            if nama_fitur is not None:
                print(
                f"model-{i+1} akurasi : {self.score_accuracy(pred, data_y_oob)} dan OOB error : {1 - self.score_accuracy(pred, data_y_oob)} , kolom root : {nama_fitur[obj.root]}"
                )
            else:
                print(
                    f"model-{i+1} akurasi : {self.score_accuracy(pred, data_y_oob)} dan OOB error : {1 - self.score_accuracy(pred, data_y_oob)}"
                )
        

    def fit_predict(self, x: np.ndarray, y: np.ndarray, nama_fitur=None):
        """
        Melakukan pembelajaran pada data x dan label y menggunakan metode Random Forest, serta mengembalikan hasil prediksi.

        Parameters:
        - x (np.ndarray): Data fitur untuk pembelajaran.
        - y (np.ndarray): Label yang sesuai dengan data fitur.
        - nama_fitur (list, optional): Nama fitur untuk keperluan informasi tambahan. Default: None.

        Returns:
        - label: Label hasil prediksi setelah melakukan pembelajaran.
        """
        self.fit(x, y, nama_fitur)

        predict = []
        for data in x:
            temp_predict = []
            for m in self.__tree:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))

        self.__label = np.array(predict)
        return self.__label
    
    def predict(self, x: np.ndarray):
        """
        Melakukan prediksi pada data x menggunakan metode Random Forest.

        Parameters:
        - x (np.ndarray): Data fitur untuk prediksi.

        Returns:
        - label: Label hasil prediksi.
        """
        predict = []
        for data in x:
            temp_predict = []
            for m in self.__tree:
                temp_predict.append(m.predict(data))
            
            predict.append(mode(temp_predict))

        self.__label = np.array(predict)
        return self.__label
    

    def score_accuracy(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Menghitung akurasi prediksi berdasarkan hasil prediksi (y_pred) dan label sebenarnya (y_true).

        Parameters:
        - y_pred (np.ndarray): Hasil prediksi.
        - y_true (np.ndarray): Label sebenarnya.

        Returns:
        - accuracy: Akurasi prediksi.
        """
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true / len(y_pred)

class Stacking:
    """
    Kelas Stacking digunakan untuk melakukan ensemble stacking pada model-model yang diberikan.

    Parameters:
    - base_model (list): List dari model-model dasar yang akan digunakan dalam stacking.
    - meta_model: Model yang akan digunakan untuk melakukan prediksi pada data hasil stacking.
    - val_size (float): Ukuran validasi yang akan digunakan dalam pembagian data train dan validation.
    - random_state: Seed untuk mengontrol randomness dalam pembagian data train dan validation.

    Attributes:
    - label: Label hasil prediksi dari model stacking.

    Methods:
    - fit(x, y): Melakukan training pada model stacking menggunakan data x dan label y.
    - fit_predict(x, y): Melakukan training pada model stacking menggunakan data x dan label y, dan mengembalikan hasil prediksi.
    - predict(x): Melakukan prediksi menggunakan model stacking pada data x.
    - score_accuracy(y_pred, y_true): Menghitung akurasi dari prediksi model stacking.

    """

    def __init__(self, base_model: list, meta_model, val_size=0.2, random_state=None):
        self.__base_model = base_model
        self.__meta_model = meta_model
        self.__label = None
        self.__val_size = val_size
        self.__random_state = random_state
    
    @property
    def label(self):
        """
        Properti label untuk mengakses hasil prediksi dari model stacking.

        Returns:
        - label: Hasil prediksi dari model stacking.

        """
        pass

    @label.getter
    def label__(self):
        return self.__label
    
    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Melakukan training pada model stacking menggunakan data x dan label y.

        Parameters:
        - x (np.ndarray): Data input untuk training.
        - y (np.ndarray): Label untuk training.

        """
        X_train, X_val, y_train, y_val = train_val_split(x, y, self.__val_size, self.__random_state)
        pred_base_model = []
        for m in self.__base_model:
            pred = m.fit_predict(X_train, y_train)
            print(f"Model {m.name} akurasi : {self.score_accuracy(pred, y_train)}")
            pred_base_model.append(m.predict(X_val))
            
        
        data_train_meta = np.column_stack(pred_base_model)
        prediksi = self.__meta_model.fit_predict(data_train_meta, y_val)
        self.__label = prediksi
    
    def fit_predict(self, x: np.ndarray, y: np.ndarray):
        """
        Melakukan training pada model stacking menggunakan data x dan label y, dan mengembalikan hasil prediksi.

        Parameters:
        - x (np.ndarray): Data input untuk training.
        - y (np.ndarray): Label untuk training.

        Returns:
        - label: Hasil prediksi dari model stacking.

        """
        self.fit(x, y)
        return self.__label
    
    def predict(self, x: np.ndarray):
        """
        Melakukan prediksi menggunakan model stacking pada data x.

        Parameters:
        - x (np.ndarray): Data input untuk prediksi.

        Returns:
        - prediksi: Hasil prediksi dari model stacking.

        """
        pred_base_model = []
        for m in self.__base_model:
            pred = m.predict(x)
            pred_base_model.append(pred)
            
        data_train_meta = np.column_stack(pred_base_model)
        prediksi = self.__meta_model.predict(data_train_meta)
        
        return prediksi
        

    def score_accuracy(self, y_pred: np.ndarray, y_true: np.ndarray):
        """
        Menghitung akurasi dari prediksi model stacking.

        Parameters:
        - y_pred (np.ndarray): Hasil prediksi dari model stacking.
        - y_true (np.ndarray): Label yang sebenarnya.

        Returns:
        - akurasi: Akurasi dari prediksi model stacking.

        """
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1
        
        return true / len(y_pred)
                

class BoostingRegressi:
    """
    Kelas BoostingRegressi digunakan untuk melakukan boosting pada model regresi.
    """

    def __init__(self, learning_rate: float = 0.2, n_estimator=10):
        """
        Inisialisasi objek BoostingRegressi.

        Parameters:
        - learning_rate (float): Tingkat pembelajaran untuk setiap model regresi. Default: 0.2.
        - n_estimator (int): Jumlah estimator (model regresi) yang akan digunakan dalam boosting. Default: 10.
        """
        self.__learning_rate = learning_rate
        self.__n_estimator = n_estimator
        self.__model = []
        self.__mean_target = None
        self.__y = None
    
    @property
    def model(self):
        """
        Properti model digunakan untuk mengakses daftar model regresi yang telah dilatih.

        Returns:
        - list: Daftar model regresi.
        """
        pass

    @model.getter
    def model_(self):
        return self.__model

    
    def fit(self, x: np.ndarray, y: np.ndarray):
        """
        Melatih model regresi dengan menggunakan data x dan y.

        Parameters:
        - x (np.ndarray): Data fitur yang akan digunakan untuk melatih model.
        - y (np.ndarray): Data target yang akan digunakan untuk melatih model.
        """
        self.__mean_target = np.mean(y)
        self.__y = y
        prediksi = np.full_like(y, np.mean(y))

        for _ in range(self.__n_estimator):
            residual = y - prediksi
            linear_model = RegresiLinear()
            linear_model.fit(x, residual)
            
            prediksi += self.__learning_rate * linear_model.predict(x)
            self.__model.append(linear_model)
    
    def fit_predict(self, x: np.ndarray, y: np.ndarray):
        """
        Melatih model regresi dengan menggunakan data x dan y, kemudian melakukan prediksi.

        Parameters:
        - x (np.ndarray): Data fitur yang akan digunakan untuk melatih model dan melakukan prediksi.
        - y (np.ndarray): Data target yang akan digunakan untuk melatih model.

        Returns:
        - np.ndarray: Hasil prediksi.
        """
        self.fit(x, y)
        pred = np.full_like(y, self.__mean_target)
        for mod in self.__model:
            pred += (self.__learning_rate * mod.predict(x))
        return pred

    def predict(self, x: np.ndarray):
        """
        Melakukan prediksi menggunakan model regresi yang telah dilatih.

        Parameters:
        - x (np.ndarray): Data fitur yang akan digunakan untuk melakukan prediksi.

        Returns:
        - np.ndarray: Hasil prediksi.
        """
        pred = np.full_like(x, shape=(x.shape[0],), fill_value=self.__mean_target)
        for mod in self.__model:
            pred += (self.__learning_rate * mod.predict(x))
        if len(x.shape) == 1:
            pred = pred[0]
        return pred
    



        











            
            
            


            





        



import numpy as np

class AlgoritmaGenetika:
    """
    Kelas yang merepresentasikan algoritma genetika.

    Atribut:
        n (int): Jumlah individu dalam populasi.
        fitness_function (function): Fungsi fitness yang digunakan untuk mengevaluasi individu.
        partisi_kromosom (int): Jumlah partisi dalam kromosom.
        isi_kromosom (dict): Nilai-nilai yang mungkin untuk setiap partisi dalam kromosom.
        iterasi (int): Jumlah maksimum iterasi yang akan dilakukan (default adalah 0, yang berarti tidak ada batasan).

    Metode:
        inisiasi_kromosom(): Menginisialisasi populasi dengan menghasilkan kromosom acak.
        fitnes_value(populasi): Menghitung nilai fitness untuk setiap individu dalam populasi.
        select_parent(populasi): Memilih induk untuk crossover berdasarkan nilai fitness mereka.
        crossover(tipe, parent): Melakukan crossover antara dua induk untuk menciptakan keturunan baru.
        mutasi(newgen): Melakukan mutasi pada sebuah kromosom.
        elistisme(populasi, new_gen): Mengaplikasikan elitisme untuk memilih individu terbaik untuk generasi berikutnya.
        fit(crossover): Menjalankan algoritma genetika untuk mencari individu terbaik dalam populasi.

    """
    def __init__(self, n: int, fitness_function, partisi_kromosom: int, isi_kromosom: dict, iterasi: int = 0):
        self.n = n
        self.fitness_function = fitness_function
        self.partisi_kromosom = partisi_kromosom
        self.isi_kromosom = isi_kromosom
        self.iterasi = iterasi

    def inisiasi_kromosom(self) -> list:
        list_kromosom = []
        while len(list_kromosom) < self.n:
            kromosom = [np.random.default_rng().choice(self.isi_kromosom[key]) for key in self.isi_kromosom.keys()]
            kromosom.append(1)
            list_kromosom.append(kromosom)
        return list_kromosom

    def fitnes_value(self, populasi) -> np.array:
        return np.array([self.fitness_function(val) for val in populasi])

    def select_parent(self, populasi) -> list:
        fit_value = self.fitnes_value(populasi)
        prob_ind = fit_value / fit_value.sum()
        cum_prob = np.cumsum(prob_ind)
        parent_indeks = []
        for _ in range(2):
            ran_angka = np.random.default_rng().random()
            parent_indeks.append(np.argmax(cum_prob > ran_angka))
        return parent_indeks

    def crossover(self, tipe, parent):
        
        parent1 = parent[0][:-1]
        parent2 = parent[1][:-1]
        if tipe == "single":
            random = np.random.default_rng().integers(self.partisi_kromosom)
            new_gen1 = parent1[:random] + parent2[random:]
            new_gen2 = parent2[:random] + parent1[random:]
        elif tipe == 'double':
            random1 = np.random.default_rng().integers(self.partisi_kromosom)
            random2 = np.random.default_rng().integers(self.partisi_kromosom)
            while random2 == random1:
                random2 = np.random.default_rng().integers(self.partisi_kromosom)
            new_gen1 = parent1[:random1] + parent2[random1:random2] + parent1[random2:]
            new_gen2 = parent2[:random1] + parent1[random1:random2] + parent2[random2:]
        new_gen1.append(1)
        new_gen2.append(1)
        return new_gen1, new_gen2

    def mutasi(self, newgen):
        key_random = np.random.default_rng().choice(list(self.isi_kromosom.keys()))
        random_replace = np.random.default_rng().choice(self.isi_kromosom[key_random])
        newgen[key_random] = random_replace
        return newgen

    def elistisme(self, populasi, new_gen):
        new_populasi = []
        for sublist in populasi:
            sublist[-1] += 1
        populasi = populasi + new_gen
        while len(populasi) > 0:
            maxi = 0
            for i in range(len(populasi[1:])):
                if self.fitness_function(populasi[i + 1]) > self.fitness_function(populasi[maxi]):
                    maxi = i + 1
                elif self.fitness_function(populasi[i + 1]) == self.fitness_function(populasi[maxi]) and populasi[i + 1][-1] < populasi[maxi][-1]:
                    maxi = i + 1
            new_populasi.append(populasi[maxi])
            populasi.pop(maxi)
        populasi = sorted(populasi, key=lambda individu: self.fitness_function(individu), reverse=True)
        new_populasi = new_populasi+populasi
        if len(new_populasi) != self.n:
            elimi = len(new_populasi) - self.n
            new_populasi = new_populasi[:-elimi]
        return new_populasi

    def fit(self, crossover="single"):
        populasi = self.inisiasi_kromosom()
        iterasi = self.iterasi
        prev_avg_fitness = None
        gen = 0
        count = 0
        if crossover == 'double' and self.partisi_kromosom <=2:
            raise Exception("Partisi kromosom harus lebih dari 2")
        while True:
            max_gen = max([self.fitness_function(individu) for individu in populasi])
            print(f"gen-{gen} - best fitnes : {max_gen}")
            parent_index = self.select_parent(populasi)
            parent = [populasi[i] for i in parent_index]
            new_gen = self.crossover(crossover, parent)
            new_gen = [self.mutasi(new_gen[i]) for i in range(2)]
            new_populasi = self.elistisme(populasi, new_gen)
            populasi = new_populasi
            avg_fitness = sum(self.fitness_function(individu) for individu in populasi) / len(populasi)
            if prev_avg_fitness is not None and abs(avg_fitness - prev_avg_fitness) < 0.001:
                break
            prev_avg_fitness = avg_fitness
           
            count += 1
            gen += 1
            if iterasi != 0 and count >= iterasi:
                break
        return populasi[0][:-1]
            
            
        
            
            
        
        
        
            
        
            
        
    


"""
Kumpulan algoritma unsupervised learning
"""
import numpy as np
import seaborn as sns
from ..matematika.distance import euclidian_distance,kuadrat_jarak

class Kmeans:
    """
    Implementasi algoritma K-Means untuk pengelompokan data tanpa supervisi.

    Parameters:
    - cluster (int): Jumlah cluster yang diinginkan. Default: 2.
    - inital_centroid (None or dict): Pusat-pusat awal untuk setiap cluster. Jika None, pusat-pusat awal akan diinisialisasi secara acak. Default: None.
    - limit_loop (int): Batas iterasi maksimum. Default: 300.
    - konvergen (None or int): Jumlah iterasi yang diperlukan untuk mencapai konvergensi. Jika None, tidak ada konvergensi yang diperlukan. Default: None.
    - random_state (int): Seed untuk menghasilkan angka acak. Default: 2.
    """

    def __init__(self, cluster: int = 2, inital_centroid=None, limit_loop: int = 300, konvergen=None, random_state: int = 2):
        self.__cluster = cluster
        self.__inital_centroid = inital_centroid
        self.__limit_loop = limit_loop
        self.__historycentroid = {}
        self.__konvergen = konvergen
        self.__random_state = random_state
        self.__cluster_fit = None
        self.__inersia = None
        self.__x = None
        self.__label = None

    @property
    def inersia(self):
        """
        Getter untuk mengakses nilai inersia dari model K-Means.

        Returns:
        - inersia (float): Nilai inersia dari model K-Means.
        """
        pass

    @property
    def label(self):
        """
        Getter untuk mengakses label hasil pengelompokan data.

        Returns:
        - label (np.ndarray): Array numpy berisi label hasil pengelompokan data.
        """
        pass

    def sethistory_centroid(self, loop, list_centroid):
        """
        Menyimpan pusat-pusat centroid pada iterasi tertentu.

        Parameters:
        - loop: Nomor iterasi.
        - list_centroid: Daftar pusat-pusat centroid pada iterasi tersebut.
        """
        self.__historycentroid[loop] = list_centroid

    def history_centroid(self) -> dict:
        """
        Mengembalikan dictionary yang berisi daftar pusat-pusat centroid pada setiap iterasi.

        Returns:
        - historycentroid (dict): Dictionary yang berisi daftar pusat-pusat centroid pada setiap iterasi.
        """
        return self.__historycentroid

    def fit_predict(self, x: np.ndarray) -> np.ndarray:
        """
        Melakukan pengelompokan data menggunakan model K-Means dan mengembalikan label hasil pengelompokan.

        Parameters:
        - x (np.ndarray): Array numpy berisi data yang akan dikelompokkan.

        Returns:
        - predict (np.ndarray): Array numpy berisi label hasil pengelompokan data.
        """
        self.__x = x
        self.fit(x)
        predict = [None for _ in range(len(x))]
        for key3 in self.__cluster_fit.keys():
            for index in self.__cluster_fit[key3]:
                predict[index] = key3
        self.__label = np.array(predict)
        return np.array(predict)

    def fit(self, x: np.ndarray):
        """
        Melakukan proses training model K-Means.

        Parameters:
        - x (np.ndarray): Array numpy berisi data yang akan digunakan untuk training.
        """
        self.__x = x
        np.random.seed(self.__random_state)
        if self.__inital_centroid == None:
            self.__inital_centroid = {}
            for i in range(self.__cluster):
                centroid = x[np.random.choice(x.shape[0], 1, replace=False)]
                self.__inital_centroid[i] = list(centroid)
        loop = 0
        konv = 0
        while loop < self.__limit_loop:
            cluster = {}
            for i in range(self.__cluster):
                cluster[i] = []
            self.sethistory_centroid(loop, self.__inital_centroid)
            for index, value in enumerate(x):
                cls = None
                jrk = None
                for key in self.__inital_centroid.keys():
                    jarak = euclidian_distance(value, np.array(self.__inital_centroid[key]))
                    if jrk == None or jarak <= jrk:
                        jrk = jarak
                        cls = key
                cluster[cls].append(index)
            update_centroid = {}

            for key2 in cluster.keys():
                if len(cluster[key2]) != 0:
                    init = x[cluster[key2][0]].copy()
                    for rec in cluster[key2][1:]:
                        init += x[rec].copy()
                    init = init / len(cluster[key2])
                else:
                    init = list(x[np.random.choice(x.shape[0], 1, replace=False)])

                update_centroid[key2] = list(init)

            if konv == self.__konvergen:
                break

            if update_centroid == self.__inital_centroid:
                konv += 1
            else:
                self.__inital_centroid = update_centroid
            loop += 1

        self.__cluster_fit = cluster

    @inersia.getter
    def inersia_(self):
        """
        Getter untuk mengakses nilai inersia dari model K-Means.

        Returns:
        - inersia (float): Nilai inersia dari model K-Means.
        """
        total = 0

        for key in self.__cluster_fit.keys():
            centroid = self.__inital_centroid[key]
            for rec in self.__cluster_fit[key]:
                total += kuadrat_jarak(self.__x[rec], np.array(centroid))
        self.__inersia = total
        return self.__inersia

    @label.getter
    def label__(self):
        """
        Getter untuk mengakses label hasil pengelompokan data.

        Returns:
        - label (np.ndarray): Array numpy berisi label hasil pengelompokan data.
        """
        return self.__label

class DBscan:
    """
    Implementasi algoritma DBSCAN (Density-Based Spatial Clustering of Applications with Noise).

    Parameters:
    - eps (float): Jarak maksimum antara dua sampel agar dianggap tetangga.
    - minpts (int): Jumlah minimum sampel dalam radius eps agar suatu sampel dianggap sebagai core point.

    Attributes:
    - label (numpy.ndarray): Label hasil klasterisasi setelah pemanggilan metode `fit_predict`.
    - jarak (list): Daftar jarak antara setiap pasangan sampel yang dihitung selama pemanggilan metode `fit`.

    Methods:
    - fit(x: np.ndarray): Melakukan klasterisasi pada data x.
    - fit_predict(x: np.ndarray) -> np.ndarray: Melakukan klasterisasi pada data x dan mengembalikan label hasil klasterisasi.
    """

    def __init__(self, eps: float, minpts: int):
        self.__eps = eps
        self.__minpts = minpts
        self.__cluster = {}
        self.__label = None
        self.__jarak = []

    @property
    def label(self):
        """
        Properti label untuk mengakses label hasil klasterisasi.

        Returns:
        - label (numpy.ndarray): Label hasil klasterisasi.
        """
        return self.__label

    @property
    def jarak(self):
        """
        Properti jarak untuk mengakses daftar jarak antara setiap pasangan sampel yang dihitung selama pemanggilan metode `fit`.

        Returns:
        - jarak (list): Daftar jarak antara setiap pasangan sampel.
        """
        return self.__jarak

    def fit(self, x: np.ndarray):
        """
        Melakukan klasterisasi pada data x.

        Parameters:
        - x (np.ndarray): Data yang akan dikelompokkan.

        Returns:
        - None
        """
        cluster = 1
        visited = []
        unvisited = [i for i in range(len(x))]
        noise = []
        N = []
        while len(unvisited) != 0:
            init = np.random.choice(unvisited, 1, replace=False)[0]
            visited.append(init)
            unvisited.remove(init)
            elemen = []
            for index, value in enumerate(x):
                if index != init:
                    jarak = euclidian_distance(x[init], value)
                    self.__jarak.append(jarak)
                    if jarak <= self.__eps:
                        N.append(index)

            if len(N) >= self.__minpts:
                elemen.append(init)
            else:
                N = []
                noise.append(init)

            while len(N) > 0:

                if N[0] not in visited:
                    visited.append(N[0])
                    unvisited.remove(N[0])
                el2 = []
                for idx2, value in enumerate(x):
                    if N[0] != idx2:
                        jarak = euclidian_distance(x[N[0]], value)
                        self.__jarak.append(jarak)
                        if jarak <= self.__eps:
                            el2.append(idx2)
                if len(el2) >= self.__minpts:
                    for j in el2:
                        if j not in visited and j not in N and j not in noise:
                            N.append(j)

                found = False
                for lst in self.__cluster.values():

                    if N[0] in lst:
                        found = True
                        break
                if not found:
                    elemen.append(N[0])

                N.pop(0)

            if len(elemen) > 0:
                self.__cluster[cluster] = elemen
                cluster += 1
        self.__cluster[-1] = noise

    def fit_predict(self, x: np.ndarray) -> np.ndarray:
        """
        Melakukan klasterisasi pada data x dan mengembalikan label hasil klasterisasi.

        Parameters:
        - x (np.ndarray): Data yang akan dikelompokkan.

        Returns:
        - label (numpy.ndarray): Label hasil klasterisasi.
        """
        self.fit(x.copy())
        predict = [None for _ in range(len(x))]
        for key in self.__cluster.keys():
            for index in self.__cluster[key]:
                predict[index] = key
        self.__label = np.array(predict)
        return self.__label

    @label.getter
    def label_(self):
        """
        Properti label_ untuk mengakses label hasil klasterisasi.

        Returns:
        - label (numpy.ndarray): Label hasil klasterisasi.
        """
        return self.__label

    @jarak.getter
    def jarak_(self):
        """
        Properti jarak_ untuk mengakses daftar jarak antara setiap pasangan sampel yang dihitung selama pemanggilan metode `fit`.

        Returns:
        - jarak (list): Daftar jarak antara setiap pasangan sampel.
        """
        return self.__jarak

    
                    

                





                

            

            




    

    



import numpy as np

class TreeNode:
    """
    Representasi dari simpul dalam pohon keputusan.

    Args:
        feature (optional): Fitur pemisah untuk simpul ini.
        value (optional): Nilai pemisah untuk simpul ini.
        result (optional): Hasil jika simpul ini adalah simpul daun.
        left (optional): Sub-pohon kiri (nilai kurang dari atau sama dengan pemisah).
        right (optional): Sub-pohon kanan (nilai lebih dari pemisah).
    """

    def __init__(self, feature=None, value=None, result=None, left=None, right=None):
        self.feature = feature
        self.value = value
        self.result = result
        self.left = left
        self.right = right

    


import numpy as np

def euclidian_distance(x: np.ndarray, y: np.ndarray) -> float:
    """
    Menghitung jarak euclidean antara dua vektor.

    Parameters:
    x (np.ndarray): Vektor pertama.
    y (np.ndarray): Vektor kedua.

    Returns:
    float: Jarak euclidean antara dua vektor.
    """
    delta = x - y
    square = delta ** 2
    return np.sqrt(np.sum(square))


def kuadrat_jarak(x:np.ndarray,y:np.ndarray)->float:
    """
    Menghitung kuadrat dari jarak Euclidean antara dua vektor.

    Parameters:
    x (np.ndarray): Vektor pertama.
    y (np.ndarray): Vektor kedua.

    Returns:
    float: Kuadrat dari jarak Euclidean antara x dan y.
    """
    return euclidian_distance(x,y)**2


import numpy as np
from ..graph.tree import TreeNode

class DecisionTree(TreeNode):
    """Kelas untuk membangun dan menggunakan pohon keputusan.

    Parameters:
    - depth (int): Kedalaman maksimum dari pohon keputusan. Default: 2

    Attributes:
    - __depth (int): Kedalaman maksimum dari pohon keputusan.
    - __tree (TreeNode): Pohon keputusan yang dibangun.
    - __label (np.ndarray): Label hasil prediksi.
    - root: Akar pohon keputusan.

    Methods:
    - calculate_gini(y): Menghitung nilai Gini impurity dari suatu set data.
    - calculate_split_info(X, y, feature, value): Menghitung nilai split info dari suatu pemisah.
    - calculate_gain_ratio(X, y, feature, value): Menghitung nilai Gain Ratio dari suatu pemisah.
    - split_data(X, y, feature, value): Memisahkan data berdasarkan suatu fitur dan nilai pemisah.
    - find_best_split(X, y, min_fitur=0, max_fitur=None): Mencari pemisah terbaik berdasarkan nilai Gain Ratio.
    - build_tree(X, y, depth=0, max_depth=None, max_feature=None, min_fitur=0): Membangun pohon keputusan secara rekursif.
    - fit(X, y, max_feature=None, min_fitur=0): Melatih model menggunakan data training.
    - fit_predict(X, y, max_feature=None, min_fitur=0): Melatih model dan melakukan prediksi pada data training.
    - predict(X): Melakukan prediksi pada data baru.
    - predict_tree(node, sample): Memprediksi label untuk sampel menggunakan pohon keputusan.
    - score_accuracy(y_pred, y_true): Menghitung akurasi prediksi.

    """

    def __init__(self, depth: int = 2):
        self.__depth = depth
        self.__tree: TreeNode = None
        self.__label: np.ndarray = None
        self.root = None

    @property
    def label(self):
        pass

    @property
    def tree(self):
        pass

    @tree.getter
    def tree__(self):
        return self.__tree

    @label.getter
    def label__(self):
        return self.__label

    def calculate_gini(self, y):
        # Menghitung nilai Gini impurity dari suatu set data
        classes, counts = np.unique(y, return_counts=True)
        probabilities = counts / len(y)
        gini = 1 - np.sum(probabilities**2)
        return gini

    def calculate_split_info(self, X: np.ndarray, y: np.ndarray, feature, value):
        # Menghitung nilai split info dari suatu pemisah
        m = len(y)
        X_left, y_left, X_right, y_right = self.split_data(X, y, feature, value)
        p_left = len(y_left) / m
        p_right = len(y_right) / m

        if p_left == 0 or p_right == 0:
            return 0
        split_info = -p_left * np.log2(p_left) - p_right * np.log2(p_right)
        return split_info

    def calculate_gain_ratio(self, X: np.ndarray, y: np.ndarray, feature, value):
        # Menghitung nilai Gain Ratio dari suatu pemisah
        gini_before = self.calculate_gini(y)
        m = len(y)

        X_left, y_left, X_right, y_right = self.split_data(X, y, feature, value)

        gini_left = self.calculate_gini(y_left)
        gini_right = self.calculate_gini(y_right)

        gini_after = (len(y_left) / m) * gini_left + (len(y_right) / m) * gini_right

        split_info = self.calculate_split_info(X, y, feature, value)

        if split_info == 0:
            return 0

        gain_ratio = (gini_before - gini_after) / split_info
        return gain_ratio

    def split_data(self, X: np.ndarray, y: np.ndarray, feature, value):
        # Memisahkan data berdasarkan suatu fitur dan nilai pemisah
        mask = X[:, feature] <= value
        X_left, y_left = X[mask], y[mask]
        X_right, y_right = X[~mask], y[~mask]
        return X_left, y_left, X_right, y_right

    def find_best_split(self, X: np.ndarray, y: np.ndarray, min_fitur=0, max_fitur=None):
        # Mencari pemisah terbaik berdasarkan nilai Gain Ratio
        m, n = X.shape
        best_gain_ratio = 0
        best_feature = None
        best_value = None
        if max_fitur == None:
            fitur = [i for i in range(n)]
        else:
            if min_fitur != 0:
                max_fitur = np.random.randint(min_fitur, max_fitur)
            fitur = np.random.choice(n, size=max_fitur, replace=False)

        for feature in fitur:
            unique_values = np.unique(X[:, feature])
            for value in unique_values:
                gain_ratio = self.calculate_gain_ratio(X, y, feature, value)

                if gain_ratio > best_gain_ratio:
                    best_gain_ratio = gain_ratio
                    best_feature = feature
                    best_value = value

        return best_feature, best_value

    def build_tree(self, X: np.ndarray, y: np.ndarray, depth=0, max_depth=None, max_feature=None, min_fitur=0):
        # Membangun pohon keputusan secara rekursif
        if depth == max_depth or len(np.unique(y)) == 1:
            # Jika mencapai kedalaman maksimum atau semua label sama, buat leaf node
            return TreeNode(result=np.bincount(y).argmax())

        best_feature, best_value = self.find_best_split(X, y, min_fitur, max_feature)
        if depth == 0:
            self.root = best_feature

        if best_feature is None:
            # Jika tidak dapat menemukan pemisah, buat leaf node
            return TreeNode(result=np.bincount(y).argmax())

        X_left, y_left, X_right, y_right = self.split_data(X, y, best_feature, best_value)

        left_subtree = self.build_tree(X_left, y_left, depth + 1, max_depth, max_feature, min_fitur)
        right_subtree = self.build_tree(X_right, y_right, depth + 1, max_depth, max_feature, min_fitur)

        return TreeNode(feature=best_feature, value=best_value, left=left_subtree, right=right_subtree)

    def fit(self, X: np.ndarray, y: np.ndarray, max_feature=None, min_fitur=0):
        """Melatih model menggunakan data training.

        Parameters:
        - X (np.ndarray): Data training.
        - y (np.ndarray): Label training.
        - max_feature (int): Jumlah fitur maksimum yang digunakan dalam mencari pemisah terbaik. Default: None
        - min_fitur (int): Jumlah fitur minimum yang digunakan dalam mencari pemisah terbaik. Default: 0

        """
        self.__tree = self.build_tree(X, y, max_depth=self.__depth, max_feature=max_feature, min_fitur=min_fitur)
        label = []
        for sam in X:
            label.append(self.predict_tree(self.__tree, sam))

        self.__label = np.array(label)

    def fit_predict(self, X: np.ndarray, y: np.ndarray, max_feature=None, min_fitur=0):
        """Melatih model dan melakukan prediksi pada data training.

        Parameters:
        - X (np.ndarray): Data training.
        - y (np.ndarray): Label training.
        - max_feature (int): Jumlah fitur maksimum yang digunakan dalam mencari pemisah terbaik. Default: None
        - min_fitur (int): Jumlah fitur minimum yang digunakan dalam mencari pemisah terbaik. Default: 0

        Returns:
        - np.ndarray: Label hasil prediksi.

        """
        self.fit(X, y, max_feature, min_fitur)
        return self.__label

    def predict(self, X: np.ndarray):
        """Melakukan prediksi pada data baru.

        Parameters:
        - X (np.ndarray): Data baru yang akan diprediksi.

        Returns:
        - np.ndarray: Label hasil prediksi.

        """
        if len(X.shape) == 1:
            return self.predict_tree(self.__tree, X)
        else:
            label = []
            for sam in X:
                label.append(self.predict_tree(self.__tree, sam))
            return np.array(label)

    def predict_tree(self, node, sample):
        """Memprediksi label untuk sampel menggunakan pohon keputusan.

        Parameters:
        - node (TreeNode): Node saat ini dalam pohon keputusan.
        - sample: Sampel yang akan diprediksi.

        Returns:
        - int: Label hasil prediksi.

        """
        if node.result is not None:
            return node.result

        if sample[node.feature] <= node.value:
            return self.predict_tree(node.left, sample)
        else:
            return self.predict_tree(node.right, sample)

    def score_accuracy(self, y_pred: np.ndarray, y_true: np.ndarray):
        """Menghitung akurasi prediksi.

        Parameters:
        - y_pred (np.ndarray): Label hasil prediksi.
        - y_true (np.ndarray): Label yang sebenarnya.

        Returns:
        - float: Akurasi prediksi.

        """
        true = 0
        for i in range(len(y_pred)):
            if y_pred[i] == y_true[i]:
                true += 1

        return true / len(y_pred)


class RegresiLinear:
    """
    Kelas RegresiLinear digunakan untuk melakukan regresi linear.
    """

    def __init__(self):
        self.__weight = 0
        self.__reg = None
    
    @property
    def reg(self):
        """
        Properti reg digunakan untuk mengakses nilai regresi yang telah dihitung.
        """
        pass

    @reg.getter
    def reg_(self):
        return self.__reg
    
    def fit(self,x:np.ndarray,y:np.ndarray):
        """
        Metode fit digunakan untuk melatih model regresi linear.

        Parameters:
            x (np.ndarray): Array 2D yang berisi fitur-fitur.
            y (np.ndarray): Array 1D yang berisi target.

        Raises:
            np.linalg.LinAlgError: Jika matriks tidak dapat diinvers.
        """
        x_T = np.transpose(x)
        multiply = x_T @ x 
        try:
            inverse = np.linalg.inv(multiply)
            self.__weight = inverse @ x_T @ y
        except np.linalg.LinAlgError:
            print("Matriks tidak dapat diinvers.")
        
    def fit_predict(self, x:np.ndarray, y:np.ndarray):
        """
        Metode fit_predict digunakan untuk melatih model dan mengembalikan hasil regresi.

        Parameters:
            x (np.ndarray): Array 2D yang berisi fitur-fitur.
            y (np.ndarray): Array 1D yang berisi target.

        Returns:
            np.ndarray: Array 1D yang berisi hasil regresi.
        """
        self.fit(x,y)
        self.__reg = x @ self.__weight
        return self.__reg
    
    def predict(self, x:np.ndarray):
        """
        Metode predict digunakan untuk melakukan prediksi menggunakan model yang telah dilatih.

        Parameters:
            x (np.ndarray): Array 2D yang berisi fitur-fitur.

        Returns:
            np.ndarray: Array 1D yang berisi hasil prediksi.
        """
        return x @ self.__weight



    



import numpy as np
import pandas as pd
from ..matematika.error import sum_error_square as sse


class KmeansGA:
    
    def __init__(self, n_clusters, n_pop,iterasi=0, random_state=2):
        self.n_clusters = n_clusters
        self.n_pop = n_pop
        self.iterasi = iterasi
        self.random_state = random_state
        
    
    def fitness_function(self,kromosom:list,x:np.array):
        
        error = 0
        for i in range(1,self.n_clusters+1):
            
            index = [idx for idx, val in enumerate(kromosom[:-1]) if val == i]
            mean = np.mean(x[index],axis=0)
            error += sse(x[index],mean)
            #print(sse(x[index],mean))
        
        return error
    
    def fitnees_value(self,populasi,x:np.array):
        
        return np.array([self.fitness_function(val,x) for val in populasi])
            
    def init_kromosom(self,partisi):
        np.random.seed(self.random_state)
        populasi = []
        while len(populasi) < self.n_pop:
            valid = False
            while not valid:
                kromosom = np.random.randint(0, self.n_clusters, size=partisi)
                # Periksa apakah setiap nilai dari 0 sampai n_clusters-1 muncul minimal sekali
                if all(val in kromosom for val in range(self.n_clusters)):
                    valid = True
            
            # Konversi ke list dan tambahkan ke populasi jika belum ada
            kromosom_list = kromosom.tolist()
            kromosom_list.append(1)
            if kromosom_list not in populasi:
                populasi.append(kromosom_list)
        
        return populasi
    
    def select_parent(self,populasi,x:np.array):
        np.random.seed(self.random_state)
        fit_value = self.fitnees_value(populasi,x)
        prob_ind = fit_value / fit_value.sum()
        cum_prob = np.cumsum(prob_ind)
        parent_indeks = []
        for _ in range(2):
            ran_angka = np.random.default_rng().random()
            parent_indeks.append(np.argmax(cum_prob > ran_angka))
        return parent_indeks
        
    
    def crossover(self, tipe, parent,partisi):
        np.random.seed(self.random_state)
        
        parent1 = parent[0][:-1]
        parent2 = parent[1][:-1]
        if tipe == "single":
            random = np.random.default_rng().integers(partisi)
            new_gen1 = parent1[:random] + parent2[random:]
            new_gen2 = parent2[:random] + parent1[random:]
        elif tipe == 'double':
            random1 = np.random.default_rng().integers(partisi)
            random2 = np.random.default_rng().integers(partisi)
            while random2 == random1:
                random2 = np.random.default_rng().integers(partisi)
            new_gen1 = parent1[:random1] + parent2[random1:random2] + parent1[random2:]
            new_gen2 = parent2[:random1] + parent1[random1:random2] + parent2[random2:]
        new_gen1.append(1)
        new_gen2.append(1)
        return new_gen1, new_gen2
    
    def mutasi(self, newgen, partisi):
        np.random.seed(self.random_state)
        key_random = np.random.default_rng().integers(partisi)
        random_replace = np.random.default_rng().integers(self.n_clusters)
        newgen[key_random] = random_replace
        return newgen
    
    def elistisme(self, populasi, new_gen,x:np.array):
        new_populasi = []
        for sublist in populasi:
            sublist[-1] += 1
        populasi = populasi + new_gen
        while len(populasi) > 0:
            mini = 0
            for i in range(len(populasi[1:])):
                if self.fitness_function(populasi[i + 1],x) < self.fitness_function(populasi[mini],x):
                    mini = i + 1
                elif self.fitness_function(populasi[i + 1],x) == self.fitness_function(populasi[mini],x) and populasi[i + 1][-1] < populasi[mini][-1]:
                    mini = i + 1
            new_populasi.append(populasi[mini])
            populasi.pop(mini)
        populasi = sorted(populasi, key=lambda individu: self.fitness_function(individu,x))
        new_populasi = new_populasi+populasi
        if len(new_populasi) != self.n_pop:
            elimi = len(new_populasi) - self.n_pop
            new_populasi = new_populasi[:-elimi]
        return new_populasi
    
    def fit(self, x:np.array,crossover="single"):
        np.random.seed(self.random_state)
        partisi = x.shape[0]
        populasi = self.init_kromosom(partisi)
        iterasi = self.iterasi
        prev_avg_fitness = None
        gen = 0
        count = 0
        if crossover == 'double' and self.partisi_kromosom <=2:
            raise Exception("Partisi kromosom harus lebih dari 2")
        while True:
            max_gen = min([self.fitness_function(individu,x) for individu in populasi])
            print(f"gen-{gen} - best fitnes : {max_gen}")
            parent_index = self.select_parent(populasi,x)
            parent = [populasi[i] for i in parent_index]
            new_gen = self.crossover(crossover, parent,partisi)
            new_gen = [self.mutasi(new_gen[i],partisi) for i in range(2)]
            new_populasi = self.elistisme(populasi, new_gen,x)
            populasi = new_populasi
            avg_fitness = sum(self.fitness_function(individu,x) for individu in populasi) / len(populasi)
            if prev_avg_fitness is not None and abs(avg_fitness - prev_avg_fitness) < 0.001:
                break
            prev_avg_fitness = avg_fitness
           
            count += 1
            gen += 1
            if iterasi != 0 and count >= iterasi:
                break
        return populasi[0][:-1]
    
    
    
        
        
        

